// file      : bbot/utility.txx -*- C++ -*-
// copyright : Copyright (c) 2014-2017 Code Synthesis Ltd
// license   : TBC; see accompanying LICENSE file

#include <iostream> // cin

#include <libbutl/fdstream.hxx>

#include <libbutl/manifest-parser.hxx>
#include <libbutl/manifest-serializer.hxx>

#include <bbot/diagnostics.hxx>

namespace bbot
{
  // run_*()
  //
  template <typename I, typename O, typename E, typename P, typename... A>
  process
  run_io_start (tracer& t,
                I&& in,
                O&& out,
                E&& err,
                const dir_path& cwd,
                const P& p,
                A&&... args)
  {
    try
    {
      return butl::process_start (
        [&t] (const char* c[], size_t n) {run_trace (t, c, n);},
        forward<I> (in),
        forward<O> (out),
        forward<E> (err),
        cwd,
        p,
        forward<A> (args)...);
    }
    catch (const process_error& e)
    {
      fail << "unable to execute " << p << ": " << e << endf;
    }
  }

  template <typename P>
  process_exit::code_type
  run_io_finish_exit (tracer&, process& pr, const P& p, bool fh)
  {
    try
    {
      pr.wait ();

      const process_exit& e (*pr.exit);

      if (e.normal ())
        return e.code ();

      diag_record dr; if (fh) dr << fail; else dr << error;
      dr << "process " << p << " terminated abnormally: "
         << e.description () << (e.core () ? " (core dumped)" : "");
    }
    catch (const process_error& e)
    {
      diag_record dr; if (fh) dr << fail; else dr << error;
      dr << "unable to execute " << p << ": " << e;
    }

    throw failed ();
  }

  template <typename P>
  inline void
  run_io_finish (tracer& t, process& pr, const P& p, bool fh)
  {
    if (run_io_finish_exit (t, pr, p, fh) == 0)
      return;

    {
      diag_record dr; if (fh) dr << fail; else dr << error;
      dr << "process " << p << " terminated with non-zero exit code";
    }

    throw failed ();
  }

  template <typename I, typename O, typename E, typename P, typename... A>
  inline process_exit::code_type
  run_io_exit (tracer& t, I&& in, O&& out, E&& err, const P& p, A&&... args)
  {
    process pr (run_io_start (t,
                              forward<I> (in),
                              forward<O> (out),
                              forward<E> (err),
                              dir_path (),
                              p,
                              forward<A> (args)...));

    return run_io_finish_exit (t, pr, p);
  }

  template <typename I, typename O, typename E, typename P, typename... A>
  inline void
  run_io (tracer& t, I&& in, O&& out, E&& err, const P& p, A&&... args)
  {
    process pr (run_io_start (t,
                              forward<I> (in),
                              forward<O> (out),
                              forward<E> (err),
                              dir_path (),
                              p,
                              forward<A> (args)...));
    run_io_finish (t, pr, p);
  }

  // curl
  //
  template <typename I, typename O, typename... A>
  http_curl::
  http_curl (tracer& t,
             I&& in,
             O&& out,
             method_type m,
             const string& url,
             A&&... options)
      : butl::curl ([&t] (const char* c[], size_t n) {run_trace (t, c, n);},
                    forward<I> (in),
                    forward<O> (out),
                    2,                // Diagnostics to stderr.
                    m,
                    url,
                    "-A", BBOT_USER_AGENT,
                    forward<A> (options)...)
  {
  }

  template <typename I, typename O, typename... A>
  tftp_curl::
  tftp_curl (tracer& t,
             I&& in,
             O&& out,
             method_type m,
             const string& url,
             A&&... options)
      : butl::curl ([&t] (const char* c[], size_t n) {run_trace (t, c, n);},
                    forward<I> (in),
                    forward<O> (out),
                    2,                // Diagnostics to stderr.
                    m,
                    url,
                    forward<A> (options)...)
  {
  }

  // *_manifest()
  //
  template <typename T>
  T
  parse_manifest (const path& f, const char* what, bool ignore_unknown)
  {
    using namespace butl;

    try
    {
      if (f.string () == "-")
        return parse_manifest<T> (std::cin, "stdin", what, ignore_unknown);

      if (!file_exists (f))
        fail << what << " manifest file " << f << " does not exist";

      ifdstream ifs (f);
      return parse_manifest<T> (ifs, f.string (), what, true, ignore_unknown);
    }
    catch (const system_error& e) // EACCES, etc.
    {
      fail << "unable to access " << what << " manifest " << f << ": " << e
           << endf;
    }
  }

  template <typename T>
  T
  parse_manifest (istream& is,
                  const string& name,
                  const char* what,
                  bool fh,
                  bool ignore_unknown)
  {
    using namespace butl;

    try
    {
      manifest_parser p (is, name);
      return T (p, ignore_unknown);
    }
    catch (const manifest_parsing& e)
    {
      diag_record dr; if (fh) dr << fail; else dr << error;

      dr << "invalid " << what << " manifest: "
         << name << ':' << e.line << ':' << e.column << ": " << e.description;
    }
    catch (const io_error& e)
    {
      diag_record dr; if (fh) dr << fail; else dr << error;

      dr << "unable to read " << what << " manifest " << name << ": " << e;
    }

    throw failed ();
  }

  template <typename T>
  void
  serialize_manifest (const T& m, const path& f, const char* what)
  {
    using namespace std;
    using namespace butl;

    try
    {
      ofdstream ofs (f, ios::binary);
      auto_rmfile arm (f); // Try to remove on failure ignoring errors.

      serialize_manifest (m, ofs, f.string (), what, true);

      ofs.close ();
      arm.cancel ();
    }
    catch (const system_error& e) // EACCES, etc.
    {
      fail << "unable to access " << what << " manifest " << f << ": " << e;
    }
  }

  template <typename T>
  void
  serialize_manifest (const T& m,
                      ostream& os,
                      const string& name,
                      const char* what,
                      bool fh)
  {
    using namespace butl;

    try
    {
      manifest_serializer s (os, name);
      m.serialize (s);
      return;
    }
    catch (const manifest_serialization& e)
    {
      diag_record dr; if (fh) dr << fail; else dr << error;

      dr << "invalid " << what << " manifest: " << e.description;
    }
    catch (const io_error& e)
    {
      diag_record dr; if (fh) dr << fail; else dr << error;

      fail << "unable to write " << what << " manifest " << name << ": " << e;
    }

    throw failed ();
  }
}
